/*=============================================================================
vitalnix/accdb - User Management Suite - Unified Account Database
useradd.c - Create a new user
  Copyright (C) Jan Engelhardt <jengelh [at] linux01 gwdg de>, 2003
  -- License restrictions apply (Dual FAL/GPL2).
  -- For details see doc/FAL.txt and doc/GPL2.txt.
=============================================================================*/
#include <ctype.h>
#include <errno.h>
#include <popt.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <libHX.h>
#include <accdb.h>
#include "sysprog/shared.h"
#define MAXFNLEN 512 /* max length for filename buffer */
#define MAXLNLEN 1024 /* max length for line buffer */

typedef unsigned char byte;

enum {
  E_SUCCESS = 0,
  E_OTHER,
  E_OPEN,
  E_NAME_USED,
  E_GID_USED,
  E_UPDATE,
  E_CLOSE,
};

static struct {
  char *lname;             // login name
  long uid;                // UID, if applicable
  char *initgrp, *suppgrp, // initial and supplemental group(s)
       *gecos,             // Gecos/comment field
       *homedir,           // Home directory
       *shell;             // Shell

  char *passwd;
  int expire, inactive;

  // useradd internal
  char *skeldir;           // skeleton directory
  int creat, dup, inter, sys;
} opt;

static struct {
  char *initgrp;        // default group
  char *homebase;       // default homebase
  char *shell;          // default shell
  int expire, inactive; // and default values for expiry and inactive fields

  // useradd internal
  char *skeldir;
} defl = {
  initgrp: "100",
  homebase: "/home",
  expire: -1,
  inactive: -1,
  skeldir: "/etc/skel",
};

extern void interactive(const void **);

static int count_args(const char **);
static int get_options(int *, const char ***);
static int only_digits(char *);
static void read_defaults(void);

static char *module_path = "*";

//-----------------------------------------------------------------------------
int main(int argc, char **argv) {
  struct accdb_module *db;
  struct accdb_user u;
  int eax, rv = E_SUCCESS;

  memset(&opt, '\0', sizeof(opt));
  read_defaults();
  opt.uid = -1;
  opt.skeldir = defl.skeldir;
  if(!(eax = get_options(&argc, (const char ***)&argv))) { return E_OTHER; }

  if((db = accdb_load(module_path)) == NULL) {
    fprintf(stderr, "Could not load ACCDB back-end module \"%s\": %s\n",
     module_path, strerror(errno));
    return E_OPEN;
  }

  opt.lname = *argv;
  if(opt.homedir == NULL && defl.homebase != NULL) {
    char buf[MAXFNLEN];
    snprintf(buf, MAXFNLEN, "%s/%s", defl.homebase, opt.lname);
    opt.homedir = HX_strdup(buf);
  }

  if(opt.inter) {
    const void *qtab[] = {
      "Login name\n[%s] > ", &opt.lname,
      "User ID (-1 for next free)\n[%d] > ", &opt.uid,
      "Initial group (or GID)\n[%s] > ", &opt.initgrp,
      "Supplemental groups (and/or GIDs), seperated by comma\n[%s] > ",
       &opt.suppgrp,
      "Comment field (real name, etc.)\n[%s] > ", &opt.gecos,
      "Home directory\n[%s] > ", &opt.homedir,
      "Skeleton dir\n[%s] > ", &opt.skeldir,
      "Shell\n[%s] > ", &opt.shell,
      // XXX password
      NULL,
    };

    p_interactive(qtab);
  }
  
  // Filling the struct for db->useradd()
  u.lname = opt.lname;
  u.uid = opt.uid;
  if(only_digits(opt.initgrp)) {
    u.gid = strtol(opt.initgrp, NULL, 0);
    u.group = NULL;
  } else {
    u.gid = -1;
    u.group = opt.initgrp;
  }
  u.gecos = opt.gecos;

  // ----------------------------------------
  struct accdb_user smask;
  memset(&smask, 0, sizeof(struct accdb_user));

  if((eax = db->open(&db->state)) <= 0) {
    fprintf(stderr, "Could not open ACCDB back-end: %s (eax=%d, errno=%d)\n",
     strerror(errno), eax, errno);
    rv = E_OPEN;
    goto __main__close_accdb;
  }

  if((eax = db->userinfo(db->state, &smask, NULL)) < 0) {
    fprintf(stderr, "Error querying the ACCDB: %s (eax=%d, errno=%d)\n",
     strerror(errno), eax, errno);
    rv = E_OTHER;
    goto __main__close_backend;
  }

  if(eax > 0) {
    fprintf(stderr, "User \"%s\" already exists\n", smask.lname);
    rv = E_NAME_USED;
  }

__main__close_backend:
  if((eax = db->close(db->state)) <= 0) {
    fprintf(stderr, "Warning: DB was not cleanly closed: %s"
     " (eax=%d, errno=%d)\n", strerror(errno), eax, errno);
    if(rv == E_SUCCESS) { rv = E_CLOSE; }
  }

__main__close_accdb:
  accdb_unload(db);
  return !1;
}

//-----------------------------------------------------------------------------
static int count_args(const char **args) {
  int argk = 0;
  while(*args++ != NULL) { ++argk; }
  return argk;
}

static int get_options(int *argc, const char ***argv) {
  static const char *_empty_argv[] = {NULL};
  const struct poptOption options_table[] = {
    // New, uudb-useradd options
    {NULL, 'I', POPT_ARG_NONE, &opt.inter, '\0',
     "Interactively prompt for parameters", NULL},
    {NULL, 'M', POPT_ARG_STRING, &module_path, '\0',
     "Use a different module than \"*\" (the default)", "name"},

    // Default options
    {NULL, 'G', POPT_ARG_STRING, &opt.suppgrp, '\0',
     "Supplementary groups", "groups"}, // UNIMPLEMENTED
    {NULL, 'c', POPT_ARG_STRING, &opt.gecos, '\0',
     "User's comment in the password file", "comment"},
    {NULL, 'd', POPT_ARG_STRING, &opt.homedir, '\0',
     "Home directory of the user", "home"},
    {NULL, 'e', POPT_ARG_STRING, &opt.expire, '\0',
     "Date when the account expires", "date"},
    {NULL, 'f', POPT_ARG_INT, &opt.inactive, '\0',
     "Days until account becomes inactive", "days"},
    {NULL, 'g', POPT_ARG_STRING, &opt.initgrp, '\0',
     "Initial group of the user", "group"},
    {NULL, 'k', POPT_ARG_STRING, &opt.skeldir, '\0',
     "Skeleton directory", "dir"},
    {NULL, 'm', POPT_ARG_NONE, &opt.creat, '\0',
     "Create home directory", NULL},
    {NULL, 'o', POPT_ARG_NONE, &opt.dup, '\0',
     "Allow creating a user with non-unique UID"
     " (might be disabled by UDB module)", NULL},
    {NULL, 'p', POPT_ARG_STRING, &opt.passwd, '\0',
     "Password", "pw"}, // UNIMPLEMENTED
    {NULL, 'r', POPT_ARG_NONE, &opt.sys, '\0',
     "System account (allow/use userid < UID_MIN)", NULL},
    {NULL, 's', POPT_ARG_STRING, &opt.shell, '\0',
     "The user's shell", "shell"},
    {NULL, 'u', POPT_ARG_LONG, &opt.uid, '\0',
     "Numerical value of the user's ID", "uid"},

    POPT_AUTOHELP
    POPT_TABLEEND
  };

  poptContext ctx;
  const char **args;
  int c, argk;

  ctx = poptGetContext(**argv, *argc, *argv, options_table, 0);
  while((c = poptGetNextOpt(ctx)) >= 0);
  if(c < -1) {
    fprintf(stderr, "%s: %s\n", poptBadOption(ctx, 0), poptStrerror(c));
    return 0;
  }

  if((args = poptGetArgs(ctx)) != NULL) {
    argk = count_args(args);
    poptDupArgv(argk, args, argc, argv);
  } else {
    *argc = 0;
    *argv = _empty_argv;
  }

  if(!opt.inter && (args == NULL || *args == NULL)) {
    // loginname is mandatory
    poptPrintUsage(ctx, stderr, 0);
    return 0;
  }

  poptFreeContext(ctx);
  return 1;
}

static int only_digits(char *p) {
  while(*p) {
    if(!isdigit(*p++)) { return 0; }
  }
  return 1;
}

static void read_defaults(void) {
  // Reads configuration options from $PATHS/default/useradd into opt
  const char *paths[] = {"etc", "/etc", NULL};
  const void *ktab[] = {
    // table of accepted options and their destination path
    "GROUP", &defl.initgrp,
    "HOME", &defl.homebase,
    "SHELL", &defl.shell,
    "EXPIRE", &defl.expire,
    "INACTIVE", &defl.inactive,
    "SKEL", &defl.skeldir,
    NULL,
  };

  const char **pathp = paths;
  char buf[MAXLNLEN];
  FILE *fp = NULL;

  while(*pathp != NULL) {
    char fnbuf[MAXFNLEN];
    snprintf(fnbuf, MAXFNLEN, "%s/default/useradd", *pathp);
    if((fp = fopen(fnbuf, "r")) != NULL) { break; }
    ++pathp;
  }

  if(fp == NULL) { return; }

  while(fgets(buf, MAXLNLEN, fp) != NULL) {
    const void **ktabp = ktab;
    char *qkey = buf, *qval;

    if(*buf == '#') { continue; }
    HX_chomp(buf);
    if((qval = strchr(buf, '=')) == NULL) { continue; }
    *qval++ = '\0';
    HX_strupper(qkey);

    while(*ktabp != NULL) {
      const char *cmp_key = (void *)*ktabp++;
      char **destp = (void *)*ktabp++;
      if(strcmp(qkey, cmp_key) == 0) { *destp = HX_strdup(qval); }
    }
  }

  fclose(fp);
  return;
}

//==[ End of file ]============================================================
